"""
This file contains collection of functions for extracting/parsing information from LLM output
"""

from .pddl_types import Action, Predicate
from collections import OrderedDict
from copy import deepcopy
import re, ast, json


def parse_params(llm_output):
    """
    Parses parameters from LLM into Python format (refer to example templates to see
    how these parameters should be formatted in LLM response).

    LLM output header should contain '### Parameters' along with structured content.
    """
    params_info = OrderedDict()
    params_heading = llm_output.split("Parameters")[1].strip().split("###")[0]
    params_str = combine_blocks(params_heading)
    params_raw = []
    for line in params_str.split("\n"):
        if line.strip() == "" or ("." not in line and not line.strip().startswith("-")):
            print(
                f"[WARNING] checking param object types - empty line or not a valid line: '{line}'"
            )
            continue
        if not (line.split(".")[0].strip().isdigit() or line.startswith("-")):
            print(f"[WARNING] checking param object types - not a valid line: '{line}'")
            continue
        try:
            params_raw.append(line.strip())
            p_info = [e for e in line.split(":")[0].split(" ") if e != ""]
            param_name, param_type = p_info[1].strip(" `"), p_info[3].strip(" `")
            params_info[param_name] = param_type
        except Exception:
            print(f"[WARNING] checking param object types - fail to parse: {line}")
            break
    return params_info, params_raw


def parse_new_predicates(llm_output) -> list[Predicate]:
    """
    Parses new predicates from LLM into Python format (refer to example templates to see
    how these predicates should be formatted in LLM response).

    LLM output header should contain '### New Predicates' along with structured content.
    """
    new_predicates = list()
    try:
        predicate_heading = (
            llm_output.split("New Predicates\n")[1].strip().split("###")[0]
        )
    except:
        raise Exception(
            "Could not find the 'New Predicates' section in the output. Provide the entire response, including all headings even if some are unchanged."
        )
    predicate_output = combine_blocks(predicate_heading)

    for p_line in predicate_output.split("\n"):
        if ("." not in p_line or not p_line.split(".")[0].strip().isdigit()) and not (
            p_line.startswith("-") or p_line.startswith("(")
        ):
            if len(p_line.strip()) > 0:
                print(f'[WARNING] unable to parse the line: "{p_line}"')
            continue
        predicate_info = p_line.split(": ")[0].strip(" 1234567890.(-)`").split(" ")
        predicate_name = predicate_info[0]
        predicate_desc = p_line.split(": ")[1].strip() if ": " in p_line else ""

        # get the predicate type info
        if len(predicate_info) > 1:
            predicate_type_info = predicate_info[1:]
            predicate_type_info = [
                l.strip(" ()`") for l in predicate_type_info if l.strip(" ()`")
            ]
        else:
            predicate_type_info = []
        params = OrderedDict()
        next_is_type = False
        upcoming_params = []

        for p in predicate_type_info:
            if next_is_type:
                if p.startswith("?"):
                    print(
                        f"[WARNING] `{p}` is not a valid type for a variable, but it is being treated as one. Should be checked by syntax check later."
                    )
                for up in upcoming_params:
                    params[up] = p
                next_is_type = False
                upcoming_params = []
            elif p == "-":
                next_is_type = True
            elif p.startswith("?"):
                upcoming_params.append(p)  # the next type will be for this variable
            else:
                print(
                    f"[WARNING] `{p}` is not corrrectly formatted. Assuming it's a variable name."
                )
                upcoming_params.append(f"?{p}")
        if next_is_type:
            print(
                f"[WARNING] The last type is not specified for `{p_line}`. Undefined are discarded."
            )
        if len(upcoming_params) > 0:
            print(
                f"[WARNING] The last {len(upcoming_params)} is not followed by a type name for {upcoming_params}. These are discarded"
            )

        # generate a clean version of the predicate
        clean = f"({predicate_name} {' '.join([f'{k} - {v}' for k, v in params.items()])}): {predicate_desc}"

        # drop the index/dot
        p_line = p_line.strip(" 1234567890.-`")
        new_predicates.append(
            {
                "name": predicate_name,
                "desc": predicate_desc,
                "raw": p_line,
                "params": params,
                "clean": clean,
            }
        )

    return new_predicates


def parse_predicates(all_predicates):
    """
    This function assumes the predicate definitions adhere to PDDL grammar.
    Assigns `params` to the predicate arguments properly. This should be run
    after retrieving a predicate list to ensure predicates are set correctly.
    """
    all_predicates = deepcopy(all_predicates)
    for i, pred in enumerate(all_predicates):
        if "params" in pred:
            continue
        pred_def = pred["raw"].split(": ")[0]
        pred_def = pred_def.strip(" ()`")  # drop any leading/strange formatting
        split_predicate = pred_def.split(" ")[1:]  # discard the predicate name
        split_predicate = [e for e in split_predicate if e != ""]

        pred["params"] = OrderedDict()
        for j, p in enumerate(split_predicate):
            if j % 3 == 0:
                assert "?" in p, f"invalid predicate definition: {pred_def}"
                assert (
                    split_predicate[j + 1] == "-"
                ), f"invalid predicate definition: {pred_def}"
                param_name, param_obj_type = p, split_predicate[j + 2]
                pred["params"][param_name] = param_obj_type
    return all_predicates


def parse_action(llm_response: str, action_name: str) -> Action:
    """
    Parse an action from a given LLM output.

    Args:
        llm_response (str): The LLM output.
        action_name (str): The name of the action.

    Returns:
        Action: The parsed action.
    """
    parameters, _ = parse_params(llm_response)
    try:
        preconditions = (
            llm_response.split("Preconditions\n")[1]
            .split("###")[0]
            .split("```")[1]
            .strip(" `\n")
        )
    except:
        raise Exception(
            "Could not find the 'Preconditions' section in the output. Provide the entire response, including all headings even if some are unchanged."
        )
    try:
        effects = (
            llm_response.split("Effects\n")[1]
            .split("###")[0]
            .split("```")[1]
            .strip(" `\n")
        )
    except:
        raise Exception(
            "Could not find the 'Effects' section in the output. Provide the entire response, including all headings even if some are unchanged."
        )
    return {
        "name": action_name,
        "parameters": parameters,
        "preconditions": preconditions,
        "effects": effects,
    }


def parse_objects(llm_response: str) -> dict[str, str]:
    """
    Extract objects from LLM response and returns dictionary string pairs object(name, type)
    Args:
        - llm_response (str):
        - types (dict[str,str]): WILL BE USED FOR CHECK ERROR RAISES
        - predicates (list[Predicate]): WILL BE USED FOR CHECK ERROR RAISES
    Returns:
        - dict[str,str]: objects
    """

    objects_head = extract_heading(llm_response, "OBJECTS")
    objects_raw = combine_blocks(objects_head)
    objects_clean = clear_comments(
        text=objects_raw, comments=[":", "//", "#", ";", "(", ")"]
    )  # Remove comments

    objects = {
        obj.split(" - ")[0].strip(" `"): obj.split(" - ")[1].strip(" `").lower()
        for obj in objects_clean.split("\n")
        if obj.strip()
    }

    return objects


def parse_initial(llm_response: str) -> list[dict[str, str]]:
    """
    Extracts state (PDDL-init) from LLM response and returns it as a list of dict strings

    Args:
        llm_response (str): The LLM output.

    Returns:
        states (list[dict[str,str]]): list of initial states in dictionaries
    """
    state_head = extract_heading(llm_response, "INITIAL")
    state_raw = combine_blocks(state_head)
    state_clean = clear_comments(state_raw)

    states = []
    for line in state_clean.split("\n"):
        line = line.strip("- `()")
        if not line:  # Skip empty lines
            continue
        name = line.split(" ")[0]
        if name == "not":
            neg = True
            name = line.split(" ")[1].strip(
                "()"
            )  # Remove the `not` and the parentheses
            params = line.split(" ")[2:]
        else:
            neg = False
            params = line.split(" ")[1:] if len(line.split(" ")) > 1 else []
        states.append({"name": name, "params": params, "neg": neg})

    return states


def parse_goal(llm_response: str) -> list[dict[str, str]]:
    """
    Extracts goal (PDDL-goal) from LLM response and returns it as a string

    Args:
        llm_response (str): The LLM output.

    Returns:
        states (list[dict[str,str]]): list of goal states in dictionaries
    """
    goal_head = extract_heading(llm_response, "GOAL")

    if goal_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the goal section of the LLM output. The goal has to be specified in a single block and as valid PDDL using the `and` and `not` operators. Likely this is caused by a too long response and limited context length. If so, try to shorten the message and exclude objects which aren't needed for the task."
        )
    goal_raw = goal_head.split("```")[1].strip()  # Only a single block in the goal
    goal_clean = clear_comments(goal_raw)

    goal_pure = (
        goal_clean.replace("and", "")
        .replace("AND", "")
        .replace("not", "")
        .replace("NOT", "")
    )
    goal = []
    for line in goal_pure.split("\n"):
        line = line.strip(" ()")
        if not line:  # Skip empty lines
            continue
        name = line.split(" ")[0]
        params = line.split(" ")[1:] if len(line.split(" ")) > 1 else []
        goal.append({"name": name, "params": params})

    return goal  # Since the goal uses `and` and `not` recombining it is difficult


def prune_types(
    types: dict[str, str], predicates: list[Predicate], actions: list[Action]
) -> dict[str, str]:
    """
    Prune types that are not used in any predicate or action.

    Args:
        types (list[str]): A list of types.
        predicates (list[Predicate]): A list of predicates.
        actions (list[Action]): A list of actions.

    Returns:
        list[str]: The pruned list of types.
    """

    used_types = {}
    for type in types:
        for pred in predicates:
            if type.split(" ")[0] in pred["params"].values():
                used_types[type] = types[type]
                break
        else:
            for action in actions:
                if type.split(" ")[0] in action["parameters"].values():
                    used_types[type] = types[type]
                    break
                if (
                    type.split(" ")[0] in action["preconditions"]
                    or type.split(" ")[0] in action["effects"]
                ):  # If the type is included in a "forall" or "exists" statement
                    used_types[type] = types[type]
                    break
    return used_types


def prune_predicates(
    predicates: list[Predicate], actions: list[Action]
) -> list[Predicate]:
    """
    Remove predicates that are not used in any action.

    Args:
        predicates (list[Predicate]): A list of predicates.
        actions (list[Action]): A list of actions.

    Returns:
        list[Predicate]: The pruned list of predicates.
    """
    used_predicates = []
    seen_predicate_names = set()

    for pred in predicates:
        for action in actions:
            # Add a space or a ")" to avoid partial matches
            names = [f"{pred['name']} ", f"{pred['name']})"]
            for name in names:
                if name in action["preconditions"] or name in action["effects"]:
                    if pred["name"] not in seen_predicate_names:
                        used_predicates.append(pred)
                        seen_predicate_names.add(pred["name"])
                    break

    return used_predicates


def extract_heading(llm_output: str, heading: str):
    """Extract the text between the heading and the next second level heading in the LLM output."""
    if heading not in llm_output:
        print("#" * 10, "LLM Output", "#" * 10)
        print(llm_output)
        print("#" * 30)
        raise ValueError(
            f"Could not find heading {heading} in the LLM output. Likely this is caused by a too long response and limited context length. If so, try to shorten the message and exclude objects which aren't needed for the task."
        )
    heading_str = (
        llm_output.split(heading)[1].split("\n### ")[0].strip()
    )  # Get the text between the heading and the next heading
    return heading_str


def convert_to_dict(llm_response: str) -> dict[str, str]:
    """Converts string into Python dictionary format."""
    dict_pattern = re.compile(
        r"{.*}", re.DOTALL
    )  # regular expression to find the JSON-like dictionary structure
    match = dict_pattern.search(
        llm_response
    )  # search for the pattern in the llm_response

    # safely evaluate the string to convert it into a Python dictionary
    if match:
        dict_str = match.group(0)
        try:
            dict = ast.literal_eval(dict_str)
            return dict
        except Exception as e:
            print(f"Error parsing dictionary: {e}")
            return None
    else:
        print("No dictionary found in the llm_response.")
        return None


def clear_comments(text: str, comments=[":", "//", "#", ";"]) -> str:
    """Remove comments from the text."""
    for comment in comments:
        text = "\n".join([line.split(comment)[0] for line in text.split("\n")])
    return text


def combine_blocks(heading_str: str):
    """Combine the inside of blocks from the heading string into a single string."""

    possible_blocks = heading_str.split("```")
    blocks = [
        possible_blocks[i] for i in range(1, len(possible_blocks), 2)
    ]  # obtain string between ```

    combined = "\n".join(blocks)

    return combined.replace(
        "\n\n", "\n"
    ).strip()  # remove leading/trailing whitespace and internal empty lines


def format_dict(dictionary):
    """Formats dictionary in JSON format easier for readability"""
    return json.dumps(dictionary, indent=4)


def format_types(type_hierarchy: dict[str, str]) -> dict[str, str]:
    """Formats Python dictionary hierarchy to PDDL formatted dictionary"""

    def process_node(node, parent_type=None):
        current_type = list(node.keys())[0]
        description = node[current_type]
        parent_type = parent_type if parent_type else current_type

        name = (
            f"{current_type} - {parent_type}"
            if current_type != parent_type
            else f"{current_type}"
        )
        desc = f"; {description}"

        result[name] = desc

        for child in node.get("children", []):
            process_node(child, current_type)

    result = {}
    process_node(type_hierarchy)
    return result


def format_predicates(predicates: list[Predicate]) -> str:
    """Formats list of predicates easier for readability"""
    if not predicates:
        return ""
    return "\n".join(
        f"{i + 1}. {pred['name']}: {pred.get('desc', 'No description provided') or 'No description provided'}"
        for i, pred in enumerate(predicates)
    )


def indent(string: str, level: int = 2):
    """Indent string helper function to format PDDL domain/task"""
    return "   " * level + string.replace("\n", f"\n{'   ' * level}")



import unittest
from l2p.task_builder import TaskBuilder
from l2p.utils import *
from .mock_llm import MockLLM


class TestTaskBuilder(unittest.TestCase):
    def setUp(self):
        self.task_builder = TaskBuilder()

    # TODO: implement test task builder functions
    def test_extract_objects(self):
        pass

    def test_extract_initial_state(self):
        pass

    def test_extract_goal_state(self):
        pass

    def test_extract_task(self):
        pass

    def test_extract_nl_conditions(self):
        pass

    def test_generate_task(self):
        pass

    def test_format_action(self):
        pass

    def test_format_objects(self):
        pass

    def test_format_initial(self):
        pass

    def test_format_goal(self):
        pass


if __name__ == "__main__":
    unittest.main()



import unittest
from l2p.domain_builder import DomainBuilder
from l2p.utils import *
from .mock_llm import MockLLM


class TestDomainBuilder(unittest.TestCase):
    def setUp(self):
        self.domain_builder = DomainBuilder()

    def test_extract_type(self):
        mock_llm_1 = MockLLM([load_file("tests/test_prompts/test_domain_builder/test_extract_type/01.txt")])
        mock_llm_2 = MockLLM([load_file("tests/test_prompts/test_domain_builder/test_extract_type/02.txt")])
        mock_llm_3 = MockLLM([load_file("tests/test_prompts/test_domain_builder/test_extract_type/03.txt")])

        types, _ = self.domain_builder.extract_type(
            model=mock_llm_1,
            domain_desc="Blocksworld is a...",
            prompt_template="Prompt template placeholder",
        )
        self.assertEqual(
            types,
            {
                "object": "Object is always root, everything is an object",
                "children": [
                    {
                        "arm": "mechanical arm that picks up and stacks blocks on other blocks or table.",
                        "children": [],
                    },
                    {
                        "block": "colored block that can be stacked or stacked on other blocks or table.",
                        "children": [],
                    },
                    {
                        "table": "surface where the blocks can be placed on top of.",
                        "children": [],
                    },
                ],
            },
        )

        with self.assertRaises(RuntimeError) as context:
            types, _ = self.domain_builder.extract_type(
                model=mock_llm_2,
                domain_desc="Blocksworld is a...",
                prompt_template="Prompt template placeholder",
            )
        self.assertIn("Max retries exceeded", str(context.exception))

    def test_extract_type_hierarchy(self):
        mock_llm_1 = MockLLM(
            [load_file("tests/test_prompts/test_domain_builder/test_extract_type_hierarchy/01.txt")]
        )
        mock_llm_2 = MockLLM(
            [load_file("tests/test_prompts/test_domain_builder/test_extract_type_hierarchy/02.txt")]
        )
        mock_llm_3 = MockLLM(
            [load_file("tests/test_prompts/test_domain_builder/test_extract_type_hierarchy/03.txt")]
        )

        expected_hierarchy = {
            "object": "Object is always root, everything is an object",
            "children": [
                {
                    "worker": "A type of object consisting of humans who do things.",
                    "children": [
                        {"administrator": "A type of worker.", "children": []},
                        {"general_worker": "A type of worker.", "children": []},
                    ],
                },
                {
                    "order": "A type of object consisting of instructions.",
                    "children": [],
                },
                {"vehicle": "A type of object consisting of vehicles.", "children": []},
                {
                    "house_component": "A type of object consisting of the components of a house.",
                    "children": [
                        {"wall": "A type of house_component.", "children": []},
                        {"floor": "A type of house_component.", "children": []},
                        {"roof": "A type of house_component.", "children": []},
                    ],
                },
                {
                    "location": "A type of object consisting of places which can be visited.",
                    "children": [
                        {
                            "house": "A type of location. ",
                            "children": [
                                {"mansion": "A type of house.", "children": []},
                                {"library": "A type of house.", "children": []},
                            ],
                        },
                        {"depot": "A type of location.", "children": []},
                    ],
                },
            ],
        }

        type_hierarchy, _ = self.domain_builder.extract_type_hierarchy(
            model=mock_llm_1,
            domain_desc="HouseConstruction is...",
            prompt_template="Prompt template placeholder",
        )
        self.assertEqual(type_hierarchy, expected_hierarchy)

        with self.assertRaises(RuntimeError) as context:
            type_hierarchy, _ = self.domain_builder.extract_type_hierarchy(
                model=mock_llm_2,
                domain_desc="HouseConstruction is...",
                prompt_template="Prompt template placeholder",
            )
        self.assertIn("Max retries exceeded", str(context.exception))

    # TODO: implement rest of test domain builder functions
    def test_extract_nl_actions(self):
        pass

    def test_extract_pddl_action(self):
        pass

    def test_extract_pddl_actions(self):
        pass

    def test_extract_parameters(self):
        pass

    def test_extract_preconditions(self):
        pass

    def test_extract_effects(self):
        pass

    def test_extract_predicates(self):
        pass

    def test_generate_domain(self):
        pass


if __name__ == "__main__":
    unittest.main()



import subprocess, os, re

# Define the exit codes
SUCCESS = 0
SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY = 1
SEARCH_PLAN_FOUND_AND_OUT_OF_TIME = 2
SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME = 3

TRANSLATE_UNSOLVABLE = 10
SEARCH_UNSOLVABLE = 11
SEARCH_UNSOLVED_INCOMPLETE = 12

TRANSLATE_OUT_OF_MEMORY = 20
TRANSLATE_OUT_OF_TIME = 21
SEARCH_OUT_OF_MEMORY = 22
SEARCH_OUT_OF_TIME = 23
SEARCH_OUT_OF_MEMORY_AND_TIME = 24

TRANSLATE_CRITICAL_ERROR = 30
TRANSLATE_INPUT_ERROR = 31
SEARCH_CRITICAL_ERROR = 32
SEARCH_INPUT_ERROR = 33
SEARCH_UNSUPPORTED = 34
DRIVER_CRITICAL_ERROR = 35
DRIVER_INPUT_ERROR = 36
DRIVER_UNSUPPORTED = 37


class FastDownward:

    def run_fast_downward(self, domain_file, problem_file, plan_file="sas_plan"):
        try:
            downward_path = "downward/fast-downward.py"

            # lmcut() = landmark-cut heuristic - refer to: https://www.fast-downward.org/PlannerUsage
            result = subprocess.run(
                [downward_path, "--alias", "lama-first", domain_file, problem_file],
                capture_output=True,
                text=True,
            )

            exitcodes = [result.returncode]

            if result.returncode == SUCCESS:
                # Planning succeeded
                with open(plan_file, "w") as f:
                    f.write(result.stdout)
                print("Planning succeeded!")
                print(
                    "All run components successfully terminated (translator: completed, search: found a plan, validate: validated a plan)"
                )

                # Extract the plan steps from the output
                plan_output = self.extract_plan_steps(result.stdout)
                if plan_output:
                    return True, plan_output
                else:
                    return False, "No plan found in the output."
            else:
                # Planning failed
                exitcode, plan_found = self.generate_portfolio_exitcode(exitcodes)
                return False, self.handle_error(exitcode, plan_found)
        except Exception as e:
            print("An error occurred while running the planner.")
            return False, str(e)

    def extract_plan_steps(self, output):
        plan_steps = re.findall(r"^\w+.*\(.*\)", output, re.MULTILINE)
        return "\n".join(plan_steps)

    def handle_error(self, exitcode, plan_found):
        if plan_found:
            if exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY:
                return "Plan found but the search ran out of memory."
            elif exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_TIME:
                return "Plan found but the search ran out of time."
            elif exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME:
                return "Plan found but the search ran out of memory and time."
            else:
                return f"Unknown plan occurred with exit code: {exitcode}"
        else:
            if exitcode == TRANSLATE_UNSOLVABLE:
                return "Translate phase determined the problem is unsolvable."
            elif exitcode == SEARCH_UNSOLVABLE:
                return "Search phase determined the problem is unsolvable."
            elif exitcode == SEARCH_UNSOLVED_INCOMPLETE:
                return "Search phase was incomplete and did not solve the problem."
            elif exitcode == TRANSLATE_OUT_OF_MEMORY:
                return "Translate phase ran out of memory."
            elif exitcode == TRANSLATE_OUT_OF_TIME:
                return "Translate phase ran out of time."
            elif exitcode == SEARCH_OUT_OF_MEMORY:
                return "Search phase ran out of memory."
            elif exitcode == SEARCH_OUT_OF_TIME:
                return "Search phase ran out of time."
            elif exitcode == SEARCH_OUT_OF_MEMORY_AND_TIME:
                return "Search phase ran out of memory and time."
            elif exitcode == TRANSLATE_CRITICAL_ERROR:
                return "Critical error in translate phase."
            elif exitcode == TRANSLATE_INPUT_ERROR:
                return "Input error in translate phase."
            elif exitcode == SEARCH_CRITICAL_ERROR:
                return "Critical error in search phase."
            elif exitcode == SEARCH_INPUT_ERROR:
                return "Input error in search phase."
            elif exitcode == SEARCH_UNSUPPORTED:
                return "Search phase encountered an unsupported feature."
            elif exitcode == DRIVER_CRITICAL_ERROR:
                return "Critical error in the driver."
            elif exitcode == DRIVER_INPUT_ERROR:
                return "Input error in the driver."
            elif exitcode == DRIVER_UNSUPPORTED:
                return "Driver encountered an unsupported feature."
            else:
                return f"Unknown error occurred with exit code: {exitcode}"

    def is_unrecoverable(self, exitcode):
        # Exit codes in the range from 30 to 39 represent unrecoverable failures.
        return 30 <= exitcode < 40

    def generate_portfolio_exitcode(self, exitcodes):

        print("Exit codes: {}".format(exitcodes))
        exitcodes = set(exitcodes)
        unrecoverable_codes = [
            code for code in exitcodes if self.is_unrecoverable(code)
        ]

        # There are unrecoverable exit codes.
        if unrecoverable_codes:
            print("Error: Unexpected exit codes: {}".format(unrecoverable_codes))
            if len(unrecoverable_codes) == 1:
                return (unrecoverable_codes[0], False)
            else:
                return (SEARCH_CRITICAL_ERROR, False)

        # At least one plan was found.
        if SUCCESS in exitcodes:
            if SEARCH_OUT_OF_MEMORY in exitcodes and SEARCH_OUT_OF_TIME in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME, True)
            elif SEARCH_OUT_OF_MEMORY in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY, True)
            elif SEARCH_OUT_OF_TIME in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_TIME, True)
            else:
                return (SUCCESS, True)

        # A config proved unsolvability or did not find a plan.
        for code in [SEARCH_UNSOLVABLE, SEARCH_UNSOLVED_INCOMPLETE]:
            if code in exitcodes:
                return (code, False)

        # No plan was found due to hitting resource limits.
        if SEARCH_OUT_OF_MEMORY in exitcodes and SEARCH_OUT_OF_TIME in exitcodes:
            return (SEARCH_OUT_OF_MEMORY_AND_TIME, False)
        elif SEARCH_OUT_OF_MEMORY in exitcodes:
            return (SEARCH_OUT_OF_MEMORY, False)
        elif SEARCH_OUT_OF_TIME in exitcodes:
            return (SEARCH_OUT_OF_TIME, False)

        assert False, "Error: Unhandled exit codes: {}".format(exitcodes)



"""
This file contains code for calling LLMs and saving raw model outputs 
Currently, this builder class contains the generic method to run any LLMs. 
It also offers extension to OpenAI and Huggingface, but can generalise to any third-party LLM store.
"""

import logging, functools
from retry import retry
from abc import ABC, abstractmethod
from typing_extensions import override

LOG: logging.Logger = logging.getLogger(__name__)


def require_llm(func):
    """
    Decorator to check if an LLM instance is provided and catch errors.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # check if LLM instance is provided
        model = kwargs.get("model", None)
        if model is None:
            # attempt tp get model from positional argument
            for arg in args:
                if isinstance(arg, LLM):
                    model = arg
                    break

        if not model:
            raise ValueError("An LLM instance must be provided to use this method.")

        # try-catch to ensure LLM is even used properly
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(
                f"An error occurred in {func.__name__}: {e}.\n You must provide an LLM engine OR proper configuration to use L2P. Refer to https://github.com/AI-Planning/l2p."
            )
            raise

    return wrapper


class LLM(ABC):
    def __init__(self, model: str, api_key: str | None = None) -> None:
        if model not in self.valid_models():
            LOG.warning(
                f"{model} is not in the valid model list for {type(self).__name__}. Valid models are: {', '.join(self.valid_models())}."
            )
        self.model: str = model
        self.api_key: str | None = api_key

    @abstractmethod
    def query(self, prompt: str) -> str:
        """
        Abstract method to query an LLM with a given prompt and return the response.

        Args:
            prompt (str): The prompt to send to the LLM
        Returns:
            str: The response from the LLM
        """
        pass

    def query_with_system_prompt(self, system_prompt: str, prompt: str) -> str:
        """
        Abstract methody to query an LLM with a given prompt and system prompt and return the response.

        Args:
            system_prompt (str): The system prompt to send to the LLM
            prompt (str): The prompt to send to the LLM
        Returns:
            str: The response from the LLM
        """
        return self.query(system_prompt + "\n" + prompt)

    def valid_models(self) -> list[str]:
        """
        List of valid model parameters, e.g., 'gpt4o-mini' for GPT
        """
        return []


class OPENAI(LLM):
    """Accessing OpenAI"""

    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        client=None,
        stop=None,
        max_tokens=4e3,
        temperature=0,
        top_p=1,
        frequency_penalty=0.0,
        presence_penalty=0.0,
        seed=0,
    ) -> None:

        # attempt to import necessary OPENAI modules
        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError(
                "The 'openai' library is required for OPENAI but is not installed. "
                "Install it using: `pip install openai`."
            )

        try:
            import tiktoken
        except ImportError:
            raise ImportError(
                "The 'tiktoken' library is required for token processing but is not installed. "
                "Install it using: `pip install tiktoken`."
            )

        # call the parent class constructor to handle model and api_key
        super().__init__(model, api_key)

        # initialize the OpenAI client or use the one provided
        self.client = client if client else OpenAI(api_key=api_key)

        # store other parameters
        self.temperature = temperature
        self.top_p = top_p
        self.freq_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.stop = stop

        self.context_length = {
            "gpt-3.5-turbo-0125": 16e3,  # 16k tokens
            "gpt-3.5-turbo-instruct": 4e3,  # 4k tokens
            "gpt-4-1106-preview": 128e3,  # 128k tokens
            "gpt-4-turbo-2024-04-09": 128e3,  # 128k tokens
            "gpt-4": 8192,  # ~8k tokens
            "gpt-4-32k": 32768,  # ~32k tokens
            "gpt-4o": 32768,  # ~32k tokens
            "gpt-4o-mini": 32768,  # ~32k tokens
        }[model]

        self.max_tokens = max_tokens if max_tokens is not None else self.context_length
        self.tok = tiktoken.get_encoding("cl100k_base")  # For GPT3.5+
        self.in_tokens = 0
        self.out_tokens = 0

    @retry(tries=2, delay=60)
    def connect_openai(
        self,
        client,
        model,
        messages,
        temperature,
        max_tokens,
        top_p,
        frequency_penalty,
        presence_penalty,
        stop,
    ):
        return client.chat.completions.create(
            model=model,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            stop=stop,
        )

    @override
    def query(
        self,
        prompt: str,
        messages=None,
        end_when_error=False,
        max_retry=5,
        est_margin=200,
    ) -> str:
        if prompt is None and messages is None:
            raise ValueError("Prompt and messages cannot both be None")
        if messages is not None:
            messages = messages
        else:
            messages = [{"role": "user", "content": prompt}]

        # calculate # of tokens to request. At most self.max_tokens, and prompt + request < self.context_length
        current_tokens = int(
            sum([len(self.tok.encode(m["content"])) for m in messages])
        )  # estimate current usage
        requested_tokens = int(
            min(self.max_tokens, self.context_length - current_tokens - est_margin)
        )  # request with safety margin
        print(
            f"Requesting {requested_tokens} tokens from {self.model} (estimated {current_tokens - est_margin} prompt tokens with a safety margin of {est_margin} tokens)"
        )
        self.in_tokens += current_tokens

        # request response
        n_retry = 0
        conn_success = False
        while not conn_success:
            n_retry += 1
            if n_retry >= max_retry:
                break
            try:
                print(f"[INFO] connecting to the LLM ({requested_tokens} tokens)...")
                response = self.connect_openai(
                    client=self.client,
                    model=self.model,
                    messages=messages,
                    temperature=self.temperature,
                    max_tokens=requested_tokens,
                    top_p=self.top_p,
                    frequency_penalty=self.freq_penalty,
                    presence_penalty=self.presence_penalty,
                    stop=self.stop,
                )
                llm_output = response.choices[
                    0
                ].message.content  # response['choices'][0]['message']['content']
                conn_success = True
            except Exception as e:
                print(f"[ERROR] LLM error: {e}")
                if end_when_error:
                    break
        if not conn_success:
            raise ConnectionError(
                f"Failed to connect to the LLM after {max_retry} retries"
            )

        response_tokens = len(self.tok.encode(llm_output))  # Estimate response tokens
        self.out_tokens += response_tokens

        return llm_output

    def get_tokens(self) -> tuple[int, int]:
        return self.in_tokens, self.out_tokens

    def reset_tokens(self):
        self.in_tokens = 0
        self.out_tokens = 0

    @override
    def valid_models(self) -> list[str]:
        """
        List of valid model parameters for OpenAI.
        Returns a list of valid OpenAI model names.
        """
        return [
            "gpt-3.5-turbo-0125",
            "gpt-3.5-turbo-instruct",
            "gpt-4-1106-preview",
            "gpt-4-turbo-2024-04-09",
            "gpt-4",
            "gpt-4-32k",
            "gpt-4o",
            "gpt-4o-mini",
        ]


class HUGGING_FACE(LLM):
    def __init__(self, model_path: str, max_tokens=4e3, temperature=0.01, top_p=0.9):
        self.model_path = model_path
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.top_p = top_p
        self.in_tokens = 0
        self.out_tokens = 0

    def _load_transformers(self):

        # attempt to import the `transformers` library
        try:
            import transformers
        except ImportError:
            raise ImportError(
                "The 'transformers' library is required for HUGGING_FACE but is not installed. "
                "Install it using: `pip install transformers`."
            )

        # attempt to import `AutoTokenizer` from `transformers`
        try:
            from transformers import AutoTokenizer
        except ImportError:
            raise ImportError(
                "The 'transformers.AutoTokenizer' module is required but is not installed properly. "
                "Ensure that the 'transformers' library is installed correctly."
            )

        # attempt to import the `torch` library
        try:
            import torch
        except ImportError:
            raise ImportError(
                "The 'torch' library is required for HUGGING_FACE but is not installed. "
                "Install it using: `pip install torch`."
            )

        try:
            # Check if the model_path is valid by trying to load it
            self.model = transformers.pipeline(
                "text-generation",
                model=self.model_path,
                model_kwargs={"torch_dtype": torch.bfloat16},
                device_map="auto",
            )
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_path)
        except OSError as e:
            # if model_path is not found, raise an error
            raise ValueError(
                f"Model path '{self.model_path}' could not be found. Please ensure the model exists."
            )
        except Exception as e:
            # catch any other exceptions and raise a generic error
            raise RuntimeError(f"An error occurred while loading the model: {str(e)}")

    # Retry decorator to handle retries on request
    @retry(tries=2, delay=60)
    def connect_huggingface(self, input, temperature, max_tokens, top_p, numSample):
        if self.model is None or self.tokenizer is None:
            self._load_transformers()

        if numSample > 1:
            responses = []
            sequences = self.model(
                input,
                do_sample=True,
                top_k=1,
                num_return_sequences=numSample,
                max_new_tokens=max_tokens,
                return_full_text=False,
                temperature=temperature,
                top_p=top_p,
                pad_token_id=self.tokenizer.eos_token_id,
            )

            for seq in sequences:
                response = seq["generated_text"]
                responses.append(response)
            return responses

        else:
            sequences = self.model(
                input,
                do_sample=True,
                num_return_sequences=1,
                max_new_tokens=max_tokens,
                return_full_text=False,
                temperature=temperature,
                top_p=top_p,
                pad_token_id=self.tokenizer.eos_token_id,
            )

            seq = sequences[0]
            response = seq["generated_text"]

            return response

    @override
    def query(self, prompt: str, numSample=1, max_retry=3, est_margin=200) -> str:
        if prompt is None:
            raise ValueError("Prompt cannot be None")

        # Estimate current usage of tokens
        current_tokens = len(self.tokenizer.encode(prompt))
        requested_tokens = min(
            self.max_tokens, self.max_tokens - current_tokens - est_margin
        )

        print(
            f"Requesting {requested_tokens} tokens from {self.model} (estimated {current_tokens - est_margin} prompt tokens with a safety margin of {est_margin} tokens)"
        )

        # Retry logic for Hugging Face request
        n_retry = 0
        conn_success = False
        while not conn_success and n_retry < max_retry:
            n_retry += 1
            try:
                print(
                    f"[INFO] Connecting to Hugging Face model ({requested_tokens} tokens)..."
                )
                llm_output = self.connect_huggingface(
                    input=prompt,
                    temperature=self.temperature,
                    max_tokens=requested_tokens,
                    top_p=self.top_p,
                    numSample=numSample,
                )
                conn_success = True
            except Exception as e:
                print(f"[ERROR] Hugging Face error: {e}")
                if n_retry >= max_retry:
                    raise ConnectionError(
                        f"Failed to connect to the Hugging Face model after {max_retry} retries"
                    )

        # Token management
        response_tokens = len(self.tokenizer.encode(llm_output))
        self.out_tokens += response_tokens
        self.in_tokens += current_tokens

        return llm_output

    def get_tokens(self) -> tuple[int, int]:
        return self.in_tokens, self.out_tokens

    def reset_tokens(self):
        self.in_tokens = 0
        self.out_tokens = 0



import os, json


def load_file(file_path):
    _, ext = os.path.splitext(file_path)
    with open(file_path, "r") as file:
        if ext == ".json":
            return json.load(file)
        else:
            return file.read().strip()



"""
This file contains collection of functions for PDDL task generation purposes
"""

from .utils import *
from .llm_builder import LLM
from .llm_builder import require_llm
import time


class TaskBuilder:
    def __init__(
        self,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
    ):
        """
        Initializes a task builder object

        Args:
            objects (dict[str,str]): current dictionary of task objects in model
            initial (list[dict[str,str]]): current initial states in model
            goal (list[dict[str,str]]): current goal states in model
        """

        self.objects = objects
        self.initial = initial
        self.goal = goal

    """Extract functions"""

    @require_llm
    def extract_objects(
        self,
        model: LLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], str]:
        """
        Extracts objects with given predicates in current model

        Args:
            model (LLM): LLM
            problem_desc (str): problem description
            domain_desc (str): domain description
            prompt_template (str): prompt template class
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): list of predicates in current model
            max_retries (int): max # of retries if failure occurs

        Returns:
            objects (dict[str,str]): dictionary of object types {name:description}
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        types_str = "\n".join(types) if types else "No types provided."

        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicate_str)
        prompt_template = prompt_template.replace("{problem_desc}", problem_desc)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # get LLM response

                # extract respective types from response
                objects = parse_objects(llm_response)

                return objects, llm_response

            except Exception as e:
                print(
                    f"Error encountered: {e}. Retrying {attempt + 1}/{max_retries}..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract objects.")

    @require_llm
    def extract_initial_state(
        self,
        model: LLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
        max_retries: int = 3,
    ) -> tuple[list[dict[str, str]], str]:
        """
        Extracts initial states with given predicates, objects, and states in current model

        Args:
            model (LLM): LLM
            problem_desc (str): problem description
            domain_desc (str): domain description
            prompt_template (str): prompt template class
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): current list of predicates in model
            objects (dict[str,str]): current dictionary of task objects in model
            initial (list[dict[str,str]]): current initial states in model
            goal (list[dict[str,str]]): current goal states in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            initial (list[dict[str,str]]): list of dictionary of initial states [{predicate,params,neg}]
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        types_str = "\n".join(types) if types else "No types provided."
        objects_str = (
            self.format_objects(objects) if objects else "No objects provided."
        )
        initial_str = (
            self.format_initial(initial) if initial else "No initial state provided."
        )
        goal_str = self.format_goal(goal) if goal else "No goal state provided."

        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicate_str)
        prompt_template = prompt_template.replace("{objects}", objects_str)
        prompt_template = prompt_template.replace("{initial_state}", initial_str)
        prompt_template = prompt_template.replace("{goal_state}", goal_str)
        prompt_template = prompt_template.replace("{problem_desc}", problem_desc)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                # extract respective types from response
                initial = parse_initial(llm_response)

                return initial, llm_response

            except Exception as e:
                print(
                    f"Error encountered: {e}. Retrying {attempt + 1}/{max_retries}..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract initial states.")

    @require_llm
    def extract_goal_state(
        self,
        model: LLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
        max_retries: int = 3,
    ) -> tuple[list[dict[str, str]], str]:
        """
        Extracts goal states with given predicates, objects, and states in current model

        Args:
            model (LLM): LLM
            problem_desc (str): problem description
            domain_desc (str): domain description
            prompt_template (str): prompt template class
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): current list of predicates in model
            objects (dict[str,str]): current dictionary of task objects in model
            initial (list[dict[str,str]]): current initial states in model
            goal (list[dict[str,str]]): current goal states in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            goal (list[dict[str,str]]): list of dictionary of goal states [{predicate,params,neg}]
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        types_str = "\n".join(types) if types else "No types provided."
        objects_str = (
            self.format_objects(objects) if objects else "No objects provided."
        )
        initial_str = (
            self.format_initial(initial) if initial else "No initial state provided."
        )
        goal_str = self.format_goal(goal) if goal else "No goal state provided."

        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicate_str)
        prompt_template = prompt_template.replace("{objects}", objects_str)
        prompt_template = prompt_template.replace("{initial_state}", initial_str)
        prompt_template = prompt_template.replace("{goal_state}", goal_str)
        prompt_template = prompt_template.replace("{problem_desc}", problem_desc)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                # extract respective types from response
                goal = parse_goal(llm_response)

                return goal, llm_response

            except Exception as e:
                print(
                    f"Error encountered: {e}. Retrying {attempt + 1}/{max_retries}..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract goal states.")

    @require_llm
    def extract_task(
        self,
        model: LLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        actions: list[Action] = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], list[dict[str, str]], list[dict[str, str]], str]:
        """
        Extracts whole task specification in current model

        Args:
            model (LLM): LLM
            problem_desc (str): problem description
            domain_desc (str): domain description
            prompt_template (str): prompt template class
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): current list of predicates in model
            actions (list[Action]): current list of Action instances in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            objects (dict[str,str]): dictionary of object types {name:description}
            initial (list[dict[str,str]]): list of dictionary of initial states [{predicate,params,neg}]
            goal (list[dict[str,str]]): list of dictionary of goal states [{predicate,params,neg}]
            llm_response (str): the raw string LLM response
        """

        model.reset_tokens()

        # replace prompt placeholders
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        types_str = "\n".join(types) if types else "No types provided."
        action_str = (
            self.format_action(actions=actions) if actions else "No actions provided."
        )

        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicate_str)
        prompt_template = prompt_template.replace("{actions}", action_str)
        prompt_template = prompt_template.replace("{problem_desc}", problem_desc)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                # extract respective types from response
                objects = parse_objects(llm_response)
                initial = parse_initial(llm_response)
                goal = parse_goal(llm_response)

                return objects, initial, goal, llm_response

            except Exception as e:
                print(
                    f"Error encountered: {e}. Retrying {attempt + 1}/{max_retries}..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract task.")

    # NOTE: This function is experimental and may be subject to change in future versions.
    @require_llm
    def extract_nl_conditions(
        self,
        model: LLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        actions: list[Action] = None,
        objects: dict[str, str] = None,
        max_retries: int = 3,
    ) -> str:
        """
        Extracts initial and goal states in natural language

        Args:
            model (LLM): LLM
            problem_desc (str): problem description
            domain_desc (str): domain description
            prompt_template (str): prompt template class
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): current list of predicates in model
            actions (list[Action]): current list of Action instances in model
            objects (dict[str,str]): current dictionary of task objects in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        types_str = "\n".join(types) if types else "No types provided."
        objects_str = (
            self.format_objects(objects) if objects else "No objects provided."
        )
        action_str = (
            self.format_action(actions=actions) if actions else "No actions provided."
        )

        prompt_template = prompt_template.replace("{problem_desc}", problem_desc)
        prompt_template = prompt_template.replace("{actions}", action_str)
        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicate_str)
        prompt_template = prompt_template.replace("{objects}", objects_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                return llm_response

            except Exception as e:
                print(
                    f"Error encountered: {e}. Retrying {attempt + 1}/{max_retries}..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract NL task states.")

    """Delete function"""

    def delete_objects(self, name):
        if self.objects is not None:
            self.objects = {
                var: type_ for var, type_ in self.objects.items() if var != name
            }

    def delete_initial_state(self, state: dict[str, str]):
        if self.initial is not None:
            self.initial = [s for s in self.initial if s != state]

    def delete_goal_state(self, state: dict[str, str]):
        if self.goal is not None:
            self.goal = [s for s in self.goal if s != state]

    """Set functions"""

    def set_objects(self, objects: dict[str, str]):
        self.set_objects = objects

    def set_initial(self, initial: dict[str, str]):
        self.set_initial = initial

    def set_goal(self, goal: str):
        self.goal = goal

    """Get functions"""

    def get_objects(self) -> dict[str, str]:
        return self.objects

    def get_initial(self) -> dict[str, str]:
        return self.initial

    def get_goal(self) -> str:
        return self.goal

    def generate_task(
        self, domain: str, problem: str, objects: str, initial: str, goal: str
    ):
        # Write problem file
        desc = "(define\n"
        desc += f"   (problem {problem})\n"
        desc += f"   (:domain {domain})\n\n"
        desc += f"   (:objects \n{indent(objects)}\n   )\n\n"
        desc += f"   (:init\n{indent(initial)}\n   )\n\n"
        desc += f"   (:goal\n{indent(goal)}\n   )\n\n"
        desc += ")"
        desc = desc.replace("AND", "and").replace("OR", "or")
        return desc

    def format_action(self, actions: list[Action]) -> str:
        desc = ""
        for action in actions:
            param_str = "\n".join(
                [f"{name} - {type}" for name, type in action["parameters"].items()]
            )  # name includes ?
            desc += f"(:action {action['name']}\n"
            desc += f"   :parameters (\n{indent(param_str,2)}\n   )\n"
            desc += f"   :precondition\n{indent(action['preconditions'],2)}\n"
            desc += f"   :effect\n{indent(action['effects'],2)}\n"
            desc += ")\n"
        return desc

    def format_objects(self, objects: dict[str, str]) -> str:
        objects = "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
        return objects

    def format_initial(self, initial_states: list[dict[str, str]]) -> str:
        inner_str = [
            f"({state['name']} {' '.join(state['params'])})" for state in initial_states
        ]  # The main part of each predicate
        full_str = [
            f"(not {inner})" if state["neg"] else inner
            for state, inner in zip(initial_states, inner_str)
        ]  # add `not` if needed
        initial_states_str = "\n".join(
            full_str
        )  # combine the states into a single string

        return initial_states_str

    def format_goal(self, goal_states: list[dict[str, str]]) -> str:
        goal_states_str = "(AND \n"

        # loop through each dictionary in the list
        for item in goal_states:
            # extract the name and parameters from the dictionary
            name = item["name"]
            params = " ".join(item["params"])
            goal_states_str += (
                f"   ({name} {params}) \n"  # append the predicate in the desired format
            )

        goal_states_str += ")"

        return goal_states_str



"""
This file uses inputted NL descriptions to generate prompts for LLM. The user does not
have to use this class, but it is generally advisable for ease of use.
"""

import os


class PromptBuilder:
    def __init__(
        self,
        role: str = None,
        technique: str = None,
        examples: list = [],
        task: str = None,
    ):
        self.role = role  # role for LLM to follow (i.e. PDDL predicate constructor)
        self.technique = technique  # prompting technique (i.e. CoT)
        self.examples = examples  # n-shot examples for LLM to follow
        self.task = task  # dynamic placeholder given information to LLM

    def set_role(self, role):
        """Sets the role for the LLM to perform task"""
        self.role = role

    def set_technique(self, technique):
        """Sets the prompting technique for LLM to perform task"""
        self.technique = technique

    def set_examples(self, example):
        """Appends a shot examples for LLM to follow"""
        self.examples.append(example)

    def set_task(self, task):
        """
        Sets a task for the LLM by providing dynamic placeholders to generate and describe domain components.

        The `task` parameter is a structured input that includes various elements to guide the LLM in understanding
        and executing the task. The task may include descriptions, types, actions, and predicates that the LLM
        will process to generate appropriate outputs.

        Here is an example of a dynamic placeholder:
        '''
        ## Domain
        {domain_desc} - A placeholder for the description of the domain, explaining the context and purpose.
        '''

        Args:
            task (str): A structured string or template containing dynamic placeholders to specify the task.
        """
        self.task = task

    def get_role(self):
        """Returns role of the prompt given"""
        return self.role

    def get_technique(self):
        """Returns prompting technique of the prompt given"""
        return self.technique

    def get_examples(self):
        """Returns list of n-examples of the prompt given"""
        return self.examples

    def get_task(self):
        """Returns dynamic placeholder task prompt"""
        return self.task

    def remove_role(self):
        """Removes role prompt"""
        self.role = None

    def remove_technique(self):
        """Removes technique prompt"""
        self.technique = None

    def remove_examples(self, idx):
        """Removes specific index of example list"""
        del self.examples[idx]

    def remove_task(self):
        """Removes dynamic placeholder task prompt"""
        self.task = None

    def generate_prompt(self):
        """Generates the whole prompt in proper format"""
        prompt = ""

        if self.role:
            prompt += f"[ROLE]: {self.role}\n\n"
            prompt += "------------------------------------------------\n"

        if self.technique:
            prompt += f"[TECHNIQUE]: {self.technique}\n\n"
            prompt += "------------------------------------------------\n"

        if len(self.examples) > 0:
            prompt += f"[EXAMPLE(S)]:\n"
            for i, example in enumerate(self.examples, 1):
                prompt += f"Example {i}:\n{example}\n\n"

            prompt += "------------------------------------------------\n"

        if self.task:
            prompt += f"[TASK]:\nHere is the task to solve:\n{self.task}\n\n"

        return prompt.strip()



"""
This file contains collection of functions for PDDL feedback generation purposes
"""

from .utils import *
from .llm_builder import LLM
from .llm_builder import require_llm
from .domain_builder import DomainBuilder
from .task_builder import TaskBuilder
from collections import OrderedDict

domain_builder = DomainBuilder()
task_builder = TaskBuilder()


class FeedbackBuilder:

    @require_llm
    def get_feedback(
        self, model: LLM, feedback_template: str, feedback_type: str, llm_response: str
    ) -> tuple[bool, str]:
        """
        This retrieves the type of feedback user requests and returns feedack message.
        feedback_type takes in either "human" "llm" or "hybrid" which it both
        """

        model.reset_tokens()

        if feedback_type.lower() == "human":
            feedback_msg = self.human_feedback(llm_response)
        elif feedback_type.lower() == "llm":
            feedback_msg = model.query(prompt=feedback_template)
        elif feedback_type.lower() == "hybrid":
            feedback_msg = model.query(prompt=feedback_template)
            response = (
                "\nORIGINAL LLM OUTPUT:\n"
                + llm_response
                + "\nFEEDBACK:\n"
                + feedback_msg
            )
            feedback_msg.replace("no feedback".lower(), "")
            feedback_msg += self.human_feedback(response)
        else:
            raise ValueError(
                "Invalid feedback_type. Expected 'human', 'llm', or 'hybrid'."
            )

        if "no feedback" in feedback_msg.lower() or len(feedback_msg.strip()) == 0:
            return True, feedback_msg

        return False, feedback_msg

    @require_llm
    def type_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        types: dict[str, str] = None,
    ) -> tuple[dict[str, str], str]:
        """Makes LLM call using feedback prompt, then parses it into type format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            types, llm_response = domain_builder.extract_type(
                model, domain_desc, prompt
            )

        return types, llm_response

    @require_llm
    def type_hierarchy_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        type_hierarchy: dict[str, str] = None,
    ) -> tuple[dict[str, str], str]:
        """Makes LLM call using feedback prompt, then parses it into type hierarchy format"""

        model.reset_tokens()

        type_str = (
            format_dict(type_hierarchy) if type_hierarchy else "No types provided."
        )

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )
            type_hierarchy, llm_response = domain_builder.extract_type_hierarchy(
                model, domain_desc, prompt
            )

        return type_hierarchy, llm_response

    @require_llm
    def nl_action_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        nl_actions: dict[str, str] = None,
        type_hierarchy: dict[str, str] = None,
    ) -> tuple[dict[str, str], str]:
        """Makes LLM call using feedback prompt, then parses it into nl_action format"""

        model.reset_tokens()

        type_str = (
            format_dict(type_hierarchy) if type_hierarchy else "No types provided."
        )
        nl_action_str = (
            format_dict(nl_actions) if nl_actions else "No actions provided."
        )

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{nl_actions}", nl_action_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )
            nl_actions, llm_response = domain_builder.extract_type_hierarchy(
                model, domain_desc, prompt
            )

        return nl_actions, llm_response

    @require_llm
    def pddl_action_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        action: Action = None,
        predicates: list[Predicate] = None,
        types: dict[str, str] = None,
    ) -> tuple[Action, list[Predicate], str]:
        """Makes LLM call using feedback prompt, then parses it into action format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        param_str = (
            ", ".join([f"{name} - {type}" for name, type in action["params"].items()])
            if action
            else "No parameters provided"
        )
        action_name = action["name"] if action else "No action name provided"
        preconditions_str = (
            action["preconditions"] if action else "No preconditions provided."
        )
        effects_str = action["effects"] if action else "No effects provided."

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{action_name}", action_name)
        feedback_template = feedback_template.replace("{parameters}", param_str)
        feedback_template = feedback_template.replace(
            "{action_preconditions}", preconditions_str
        )
        feedback_template = feedback_template.replace("{action_effects}", effects_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            action, predicates, llm_response = domain_builder.extract_pddl_action(
                model, domain_desc, prompt, action_name
            )
        return action, predicates, llm_response

    @require_llm
    def parameter_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        parameter: OrderedDict = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] = None,
    ) -> tuple[OrderedDict, OrderedDict, str]:
        """Makes LLM call using feedback prompt, then parses it into parameter format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        param_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )
        action_name = action_name if action_name else "No action name provided."
        action_desc = action_desc if action_desc else "No action description provided."

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{action_name}", action_name)
        feedback_template = feedback_template.replace("{action_desc}", action_desc)
        feedback_template = feedback_template.replace("{parameters}", param_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            param, param_raw, llm_response = domain_builder.extract_parameters(
                model, domain_desc, prompt, action_name, action_desc, types
            )
        return param, param_raw, llm_response

    @require_llm
    def precondition_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        parameter: OrderedDict = None,
        preconditions: str = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
    ) -> tuple[str, list[Predicate], str]:
        """Makes LLM call using feedback prompt, then parses it into precondition format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        param_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )
        action_name = action_name if action_name else "No action name provided."
        action_desc = action_desc if action_desc else "No action description provided."
        precondition_str = (
            preconditions if preconditions else "No preconditions provided."
        )

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{action_name}", action_name)
        feedback_template = feedback_template.replace("{action_desc}", action_desc)
        feedback_template = feedback_template.replace("{parameters}", param_str)
        feedback_template = feedback_template.replace(
            "{action_preconditions}", precondition_str
        )

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            preconditions, new_predicates, llm_response = (
                domain_builder.extract_preconditions(
                    model, domain_desc, prompt, action_name, action_desc
                )
            )
        return preconditions, new_predicates, llm_response

    @require_llm
    def effect_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        parameter: OrderedDict = None,
        preconditions: str = None,
        effects: str = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
    ) -> tuple[str, list[Predicate], str]:
        """Makes LLM call using feedback prompt, then parses it into effects format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        param_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )
        action_name = action_name if action_name else "No action name provided."
        action_desc = action_desc if action_desc else "No action description provided."
        precondition_str = (
            preconditions if preconditions else "No preconditions provided."
        )
        effect_str = effects if effects else "No effects provided."

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{action_name}", action_name)
        feedback_template = feedback_template.replace("{action_desc}", action_desc)
        feedback_template = feedback_template.replace("{parameters}", param_str)
        feedback_template = feedback_template.replace(
            "{action_preconditions}", precondition_str
        )
        feedback_template = feedback_template.replace("{action_effects}", effect_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            effects, new_predicates, llm_response = domain_builder.extract_effects(
                model, domain_desc, prompt, action_name, action_desc
            )
        return effects, new_predicates, llm_response

    @require_llm
    def predicate_feedback(
        self,
        model: LLM,
        domain_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        nl_actions: dict[str, str] = None,
    ) -> tuple[list[Predicate], str]:
        """Makes LLM call using feedback prompt, then parses it into predicates format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        nl_action_str = (
            format_dict(nl_actions) if nl_actions else "No actions provided."
        )

        feedback_template = feedback_template.replace("{domain_desc}", domain_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{nl_actions}", nl_action_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            new_predicates, llm_response = domain_builder.extract_predicates(
                model, domain_desc, prompt
            )
        return new_predicates, llm_response

    @require_llm
    def task_feedback(
        self,
        model: LLM,
        problem_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        predicates: list[Predicate] = None,
        types: dict[str, str] = None,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
    ) -> tuple[dict[str, str], list[dict[str, str]], list[dict[str, str]], str]:
        """Makes LLM call using feedback prompt, then parses it into object, initial, and goal format"""

        model.reset_tokens()

        type_str = format_dict(types) if types else "No types provided."
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        objects_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        initial_state_str = (
            task_builder.format_initial(initial)
            if initial
            else "No initial state provided."
        )
        goal_state_str = (
            task_builder.format_goal(goal) if goal else "No goal state provided."
        )

        feedback_template = feedback_template.replace("{problem_desc}", problem_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{objects}", objects_str)
        feedback_template = feedback_template.replace(
            "{initial_state}", initial_state_str
        )
        feedback_template = feedback_template.replace("{goal_state}", goal_state_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            objects, initial, goal, _ = task_builder.extract_task(
                model, problem_desc, prompt
            )

        return objects, initial, goal, fb_msg

    @require_llm
    def objects_feedback(
        self,
        model: LLM,
        problem_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        type_hierarchy: dict[str, str] = None,
        predicates: list[Predicate] = None,
        objects: dict[str, str] = None,
    ) -> tuple[dict[str, str], str]:
        """Makes LLM call using feedback prompt, then parses it into objects format"""

        model.reset_tokens()

        type_str = (
            format_dict(type_hierarchy) if type_hierarchy else "No types provided."
        )
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        objects_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )

        feedback_template = feedback_template.replace("{problem_desc}", problem_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{objects}", objects_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            objects, llm_response = task_builder.extract_objects(
                model, problem_desc, prompt
            )

        return objects, llm_response

    @require_llm
    def initial_state_feedback(
        self,
        model: LLM,
        problem_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        type_hierarchy: dict[str, str] = None,
        predicates: list[Predicate] = None,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
    ) -> tuple[list[dict[str, str]], str]:
        """Makes LLM call using feedback prompt, then parses it into initial states format"""

        model.reset_tokens()

        type_str = (
            format_dict(type_hierarchy) if type_hierarchy else "No types provided."
        )
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        objects_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        initial_state_str = (
            task_builder.format_initial(initial)
            if initial
            else "No initial state provided."
        )

        feedback_template = feedback_template.replace("{problem_desc}", problem_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{objects}", objects_str)
        feedback_template = feedback_template.replace(
            "{initial_state}", initial_state_str
        )

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            initial, llm_response = task_builder.extract_initial_state(
                model, problem_desc, prompt
            )

        return initial, llm_response

    @require_llm
    def goal_state_feedback(
        self,
        model: LLM,
        problem_desc: str,
        llm_response: str,
        feedback_template: str,
        feedback_type: str = "llm",
        type_hierarchy: dict[str, str] = None,
        predicates: list[Predicate] = None,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
    ) -> tuple[list[dict[str, str]], str]:
        """Makes LLM call using feedback prompt, then parses it into goal states format"""

        model.reset_tokens()

        type_str = (
            format_dict(type_hierarchy) if type_hierarchy else "No types provided."
        )
        predicate_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        objects_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        initial_state_str = (
            task_builder.format_initial(initial)
            if initial
            else "No initial state provided."
        )
        goal_state_str = (
            task_builder.format_goal(goal) if goal else "No goal state provided."
        )

        feedback_template = feedback_template.replace("{problem_desc}", problem_desc)
        feedback_template = feedback_template.replace("{llm_response}", llm_response)
        feedback_template = feedback_template.replace("{types}", type_str)
        feedback_template = feedback_template.replace("{predicates}", predicate_str)
        feedback_template = feedback_template.replace("{objects}", objects_str)
        feedback_template = feedback_template.replace(
            "{initial_state}", initial_state_str
        )
        feedback_template = feedback_template.replace("{initial_state}", goal_state_str)

        no_fb, fb_msg = self.get_feedback(
            model, feedback_template, feedback_type, llm_response
        )

        if not no_fb:
            prompt = (
                f"\n\nYou now are revising your answer using feedback. Here is the feedback you outputted:\n{fb_msg}"
                f"\n\nFollow the same syntax format as the original output in your answer:\n{llm_response}"
            )

            goal, llm_response = task_builder.extract_goal_state(
                model, problem_desc, prompt
            )

        return goal, llm_response

    def human_feedback(self, info: str):
        """This enables human-in-the-loop feedback mechanism"""

        print("START OF INFO\n", info)
        print("\nEND OF INFO\n\n")
        contents = []
        print("Provide feedback (or type 'done' to finish):\n")
        while True:
            line = input()
            if line.strip().lower() == "done":
                break
            contents.append(line)
        resp = "\n".join(contents)

        if resp.strip().lower() == "no feedback":
            return "no feedback"

        return resp



"""
This file contains collection of functions for PDDL domain generation purposes
"""

from .utils import *
from .llm_builder import LLM
from .llm_builder import require_llm
from collections import OrderedDict
import re, time


class DomainBuilder:
    def __init__(
        self,
        types: dict[str, str] = None,
        type_hierarchy: dict[str, str] = None,
        predicates: list[Predicate] = None,
        nl_actions: dict[str, str] = None,
        pddl_actions: list[Action] = None,
    ):
        """
        Initializes a domain builder object

        Args:
            types (dict[str,str]): types dictionary with name: description key-value pair
            type_hierarchy (dict[str,str]): type hierarchy dictionary
            predicates (list[Predicate]): list of Predicate objects
            nl_actions (dict[str,str]): dictionary of extracted actions, where the keys are action names and values are action descriptions
            pddl_actions (list[Action]): list of Action objects
        """
        self.types = types
        self.type_hierarchy = type_hierarchy
        self.predicates = predicates
        self.nl_actions = nl_actions
        self.pddl_actions = pddl_actions

    """Extract functions"""

    @require_llm
    def extract_type(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], str]:
        """
        Extracts types with domain given

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            types (dict[str,str]): current types in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            type_dict (dict[str,str]): dictionary of types with (name:description) pair
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{types}", types_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # prompt model

                # extract respective types from response
                types = convert_to_dict(llm_response=llm_response)

                if types is not None:
                    return types, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract types.")

    @require_llm
    def extract_type_hierarchy(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], str]:
        """
        Extracts type hierarchy from types list and domain given

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            types (dict[str,str]): current types in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            type_hierarchy (dict[str,str]): dictionary of type hierarchy
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{types}", types_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:

                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                # extract respective types from response
                type_hierarchy = convert_to_dict(llm_response=llm_response)

                if type_hierarchy is not None:
                    return type_hierarchy, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract type hierarchy.")

    @require_llm
    def extract_nl_actions(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        nl_actions: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], str]:
        """
        Extract actions in natural language given domain description using LLM.

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            types (dict[str,str]): current types in model
            nl_actions (dict[str, str]): NL actions currently in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            nl_actions (dict[str, str]): a dictionary of extracted actions {action name: action description}
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."
        nl_actions_str = (
            "\n".join(f"{name}: {desc}" for name, desc in nl_actions.items())
            if nl_actions
            else "No actions provided."
        )

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{nl_actions}", nl_actions_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:

                model.reset_tokens()

                llm_response = model.query(
                    prompt=prompt_template
                )  # get LLM llm_response

                # extract respective types from response
                nl_actions = convert_to_dict(llm_response=llm_response)

                if nl_actions is not None:
                    return nl_actions, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract NL actions.")

    @require_llm
    def extract_pddl_action(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        action_name: str,
        action_desc: str = None,
        action_list: dict[str, str] = None,
        predicates: list[Predicate] = None,
        types: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[Action, list[Predicate], str]:
        """
        Extract an action and predicates from a given action description using LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): action construction prompt
            action_name (str): action name
            action_desc (str): action description
            action_list (dict[str,str]): dictionary of other actions to be translated
            predicates (list[Predicate]): list of predicates in current model
            types (dict[str,str]): current types in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            action (Action): constructed action class
            new_predicates (list[Predicate]): a list of new predicates
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."
        predicates_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        action_list_str = (
            str(action_list) if action_list else "No other actions provided"
        )

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{action_list}", action_list_str)
        prompt_template = prompt_template.replace("{action_name}", action_name)
        prompt_template = prompt_template.replace(
            "{action_desc}", action_desc if action_desc else "No description available."
        )
        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicates_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:

                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)

                # extract respective types from response
                action = parse_action(
                    llm_response=llm_response, action_name=action_name
                )
                new_predicates = parse_new_predicates(llm_response)

                if action is not None and new_predicates is not None:
                    return action, new_predicates, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract PDDL action.")

    @require_llm
    def extract_pddl_actions(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        nl_actions: dict[str, str] = None,
        predicates: list[Predicate] = None,
        types: dict[str, str] = None,
    ) -> tuple[list[Action], list[Predicate], str]:
        """
        Extract all actions from a given action description using LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): action construction prompt
            nl_actions (dict[str, str]): NL actions currently in model
            predicates (list[Predicate]): list of predicates
            types (dict[str,str]): current types in model

        Returns:
            action (Action): constructed action class
            new_predicates (list[Predicate]): a list of new predicates
            llm_response (str): the raw string LLM response
        """

        model.reset_tokens()

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."
        predicates_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        nl_actions_str = (
            format_dict(nl_actions) if nl_actions else "No actions provided."
        )

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicates_str)
        prompt_template = prompt_template.replace("{nl_actions}", nl_actions_str)

        llm_response = model.query(prompt=prompt_template)

        # extract respective types from response
        raw_actions = llm_response.split("## NEXT ACTION")

        actions = []
        for i in raw_actions:
            # define the regex patterns
            action_pattern = re.compile(r"\[([^\]]+)\]")
            rest_of_string_pattern = re.compile(r"\[([^\]]+)\](.*)", re.DOTALL)

            # search for the action name
            action_match = action_pattern.search(i)
            action_name = action_match.group(1) if action_match else None

            # extract the rest of the string
            rest_match = rest_of_string_pattern.search(i)
            rest_of_string = rest_match.group(2).strip() if rest_match else None

            actions.append(
                parse_action(llm_response=rest_of_string, action_name=action_name)
            )

        # if user queries predicate creation via LLM
        try:
            new_predicates = parse_new_predicates(llm_response)

            if predicates:
                new_predicates = [
                    pred
                    for pred in new_predicates
                    if pred["name"] not in [p["name"] for p in predicates]
                ]  # remove re-defined predicates
        except Exception as e:
            # Log or print the exception if needed
            print(f"No new predicates: {e}")
            new_predicates = None

        return actions, new_predicates, llm_response

    @require_llm
    def extract_parameters(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        action_name: str,
        action_desc: str,
        types: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[OrderedDict, list, str]:
        """
        Extracts parameters from single action description via LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            action_name (str): action name
            action_desc (str): action description
            types (dict[str,str]): current types in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            param (OrderedDict): ordered list of parameters
            param_raw (list()): list of raw parameters
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{action_name}", action_name)
        prompt_template = prompt_template.replace("{action_desc}", action_desc)
        prompt_template = prompt_template.replace("{types}", types_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # get LLM response

                # extract respective types from response
                param, param_raw = parse_params(llm_output=llm_response)

                return param, param_raw, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract parameters.")

    @require_llm
    def extract_preconditions(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        action_name: str,
        action_desc: str,
        params: list[str] = None,
        predicates: list[Predicate] = None,
        max_retries: int = 3,
    ) -> tuple[str, list[Predicate], str]:
        """
        Extracts preconditions from single action description via LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            action_name (str): action name
            action_desc (str): action description
            params (list[str]): list of parameters from action
            predicates (list[Predicate]): list of current predicates in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            preconditions (str): PDDL format of preconditions
            new_predicates (list[Predicate]): a list of new predicates
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicates_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        params_str = "\n".join(params) if params else "No parameters provided."

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{action_name}", action_name)
        prompt_template = prompt_template.replace("{action_desc}", action_desc)
        prompt_template = prompt_template.replace("{parameters}", params_str)
        prompt_template = prompt_template.replace("{predicates}", predicates_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # get LLM response

                # extract respective types from response
                preconditions = (
                    llm_response.split("Preconditions\n")[1]
                    .split("##")[0]
                    .split("```")[1]
                    .strip(" `\n")
                )
                new_predicates = parse_new_predicates(llm_output=llm_response)

                return preconditions, new_predicates, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract preconditions.")

    @require_llm
    def extract_effects(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        action_name: str,
        action_desc: str,
        params: list[str] = None,
        precondition: str = None,
        predicates: list[Predicate] = None,
        max_retries: int = 3,
    ) -> tuple[str, list[Predicate], str]:
        """
        Extracts effects from single action description via LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            action_name (str): action name
            action_desc (str): action description
            params (list[str]): list of parameters from action
            precondition (str): PDDL format of preconditions
            predicates (list[Predicate]): list of current predicates in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            effects (str): PDDL format of effects
            new_predicates (list[Predicate]): a list of new predicates
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        predicates_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        params_str = "\n".join(params) if params else "No parameters provided."

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{action_name}", action_name)
        prompt_template = prompt_template.replace("{action_desc}", action_desc)
        prompt_template = prompt_template.replace("{parameters}", params_str)
        prompt_template = prompt_template.replace("{preconditions}", precondition)
        prompt_template = prompt_template.replace("{predicates}", predicates_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # get LLM response

                # extract respective types from response
                effects = (
                    llm_response.split("Effects\n")[1]
                    .split("##")[0]
                    .split("```")[1]
                    .strip(" `\n")
                )
                new_predicates = parse_new_predicates(llm_output=llm_response)

                return effects, new_predicates, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract effects.")

    @require_llm
    def extract_predicates(
        self,
        model: LLM,
        domain_desc: str,
        prompt_template: str,
        types: dict[str, str] = None,
        predicates: list[Predicate] = None,
        nl_actions: dict[str, str] = None,
        max_retries: int = 3,
    ) -> tuple[list[Predicate], str]:
        """
        Extracts predicates via LLM

        Args:
            model (LLM): LLM
            domain_desc (str): domain description
            prompt_template (str): prompt template
            types (dict[str,str]): current types in model
            predicates (list[Predicate]): list of current predicates in model
            nl_actions (dict[str, str]): NL actions currently in model
            max_retries (int): max # of retries if failure occurs

        Returns:
            new_predicates (list[Predicate]): a list of new predicates
            llm_response (str): the raw string LLM response
        """

        # replace prompt placeholders
        types_str = format_dict(types) if types else "No types provided."
        predicates_str = (
            format_predicates(predicates) if predicates else "No predicates provided."
        )
        nl_actions_str = (
            "\n".join(f"{name}: {desc}" for name, desc in nl_actions.items())
            if nl_actions
            else "No actions provided."
        )

        prompt_template = prompt_template.replace("{domain_desc}", domain_desc)
        prompt_template = prompt_template.replace("{types}", types_str)
        prompt_template = prompt_template.replace("{predicates}", predicates_str)
        prompt_template = prompt_template.replace("{nl_actions}", nl_actions_str)

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()

                llm_response = model.query(prompt=prompt_template)  # prompt model

                # extract respective types from response
                new_predicates = parse_new_predicates(llm_output=llm_response)

                return new_predicates, llm_response

            except Exception as e:
                print(
                    f"Error encountered during attempt {attempt + 1}/{max_retries}: {e}. "
                    f"LLM Output: \n\n{llm_response if 'llm_response' in locals() else 'None'}\n\n Retrying..."
                )
                time.sleep(1)  # add a delay before retrying

        raise RuntimeError("Max retries exceeded. Failed to extract predicates.")

    """Delete functions"""

    def delete_type(self, name: str):
        """Deletes specific type from current model"""
        if self.types is not None:
            self.types = {
                type_: desc for type_, desc in self.types.items() if type_ != name
            }

    def delete_nl_action(self, name: str):
        """Deletes specific NL action from current model"""
        if self.nl_actions is not None:
            self.nl_actions = {
                action_name: action_desc
                for action_name, action_desc in self.nl_actions.items()
                if action_name != name
            }

    def delete_pddl_action(self, name: str):
        """Deletes specific PDDL action from current model"""
        if self.pddl_actions is not None:
            self.pddl_actions = [
                action for action in self.pddl_actions if action["name"] != name
            ]

    def delete_predicate(self, name: str):
        """Deletes specific predicate from current model"""
        if self.predicates is not None:
            self.predicates = [
                predicate for predicate in self.predicates if predicate["name"] != name
            ]

    """Set functions"""

    def set_types(self, types: dict[str, str]):
        """Sets types for current model"""
        self.types = types

    def set_type_hierarchy(self, type_hierarchy: dict[str, str]):
        """Sets type hierarchy for current model"""
        self.type_hierarchy = type_hierarchy

    def set_nl_actions(self, nl_actions: dict[str, str]):
        """Sets NL actions for current model"""
        self.nl_actions = nl_actions

    def set_pddl_action(self, pddl_action: Action):
        """Appends a PDDL action for current model"""
        self.pddl_actions.append(pddl_action)

    def set_predicate(self, predicate: Predicate):
        """Appends a predicate for current model"""
        self.predicates.append(predicate)

    """Get functions"""

    def get_types(self):
        """Returns types from current model"""
        return self.types

    def get_type_hierarchy(self):
        """Returns type hierarchy from current model"""
        return self.type_hierarchy

    def get_nl_actions(self):
        """Returns natural language actions from current model"""
        return self.nl_actions

    def get_pddl_actions(self):
        """Returns PDDL actions from current model"""
        return self.pddl_actions

    def get_predicates(self):
        """Returns predicates from current model"""
        return self.predicates

    def generate_domain(
        self,
        domain: str,
        types: str,
        predicates: str,
        actions: list[Action],
        requirements: list[str],
    ) -> str:
        """
        Generates PDDL domain from given information

        Args:
            domain (str): domain name
            types (str): domain types
            predicates (str): domain predicates
            actions (list[Action]): domain actions
            requirements (list[str]): domain requirements

        Returns:
            desc (str): PDDL domain
        """
        desc = ""
        desc += f"(define (domain {domain})\n"
        desc += indent(string=f"(:requirements\n  {' '.join(requirements)}\n)\n")
        desc += f"   (:types \n{indent(types)}\n   )\n\n"
        desc += f"   (:predicates \n{indent(predicates)}\n   )"
        desc += self.action_descs(actions)
        desc += "\n)"
        desc = desc.replace("AND", "and").replace(
            "OR", "or"
        )  # The python PDDL package can't handle capital AND and OR
        return desc

    def action_desc(self, action: Action) -> str:
        """Helper function to format individual action descriptions"""
        param_str = "\n".join(
            [f"{name} - {type}" for name, type in action["parameters"].items()]
        )  # name includes ?
        desc = f"(:action {action['name']}\n"
        desc += f"   :parameters (\n{param_str}\n   )\n"
        desc += f"   :precondition\n{action['preconditions']}\n"
        desc += f"   :effect\n{action['effects']}\n"
        desc += ")"
        return desc

    def action_descs(self, actions) -> str:
        """Helper function to combine all action descriptions"""
        desc = ""
        for action in actions:
            desc += "\n\n" + self.action_desc(action)
        return desc

    def format_predicates(self, predicates: list[Predicate]) -> str:
        """Helper function that formats predicate list into string"""
        return "\n".join([pred["clean"].replace(":", " ; ") for pred in predicates])



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="l2p",
    version="0.1.8",
    packages=find_packages(exclude=["tests"]),
    description="Library to connect LLMs and planning tasks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Marcus Tantakoun, Christian Muise",
    author_email="mtantakoun@gmail.com, christian.muise@gmail.com",
    install_requires=["retry", "pddl", "typing_extensions"],
    license="MIT",
    url="https://github.com/AI-Planning/l2p",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



"""
This file contains classes of PDDL types
"""

from typing import TypedDict, NewType, List, Optional
from collections import OrderedDict
from dataclasses import dataclass

ParameterList = NewType(
    "ParameterList", OrderedDict[str, str]
)  # {param_name: param_type}
ObjectList = NewType("ObjectList", dict[str, str])  # {obj_name: obj_type}


class Predicate(TypedDict):
    name: str
    desc: Optional[str]
    raw: str
    params: ParameterList
    clean: str


class Action(TypedDict):
    name: str
    raw: str
    params: ParameterList
    preconditions: str
    effects: str


# Domain details data class including predicates and actions
@dataclass
class DomainDetails:
    name: str
    domain_desc: str
    domain_pddl: str
    types: str
    requirements: list[str]
    predicates: List[Predicate]  # List of Predicate objects
    actions: List[Action]  # List of Action objects


# Problem details data class
@dataclass
class ProblemDetails:
    name: str
    problem_desc: str
    problem_pddl: str
    objects: tuple[dict[str, str], str]
    initial: tuple[dict[str, str], str]
    goal: tuple[dict[str, str], str]


# Plan details data class
@dataclass
class PlanDetails:
    plan_pddl: str
    plan_nl: str



"""
This file contains collection of functions PDDL syntax validations
"""

from collections import OrderedDict
from .pddl_parser import parse_params, parse_new_predicates, parse_predicates
from .pddl_types import Predicate


class SyntaxValidator:

    # PARAMETER CHECKS

    def validate_params(
        self, parameters: OrderedDict, types: dict[str, str]
    ) -> tuple[bool, str]:
        """Checks whether a PDDL action parameter contains types found in object types."""

        for param_name in parameters:
            param_type = parameters[param_name]

            if not any(param_type in t for t in types.keys()):
                feedback_msg = f'There is an invalid object type `{param_type}` for the parameter {param_name} not found in the types {types.keys()}. If you need to use a new type, you can emulate it with an "is_{{type}} ?o - object" precondition. Please revise the PDDL model to fix this error.'
                return False, feedback_msg

        feedback_msg = "PASS: All parameter types found in object types."
        return True, feedback_msg

    # PREDICATE CHECKS

    def validate_types_predicates(
        self, predicates: list[dict], types: dict[str, str]
    ) -> tuple[bool, str]:
        """Check if predicate name is found within any type definitions"""

        invalid_predicates = list()
        for pred in predicates:
            pred_name = pred["name"].lower()

            for type_key in types.keys():
                # extract the actual type name, disregarding hierarchical or descriptive parts
                type_name = type_key.split(" - ")[0].strip().lower()

                # check if the predicate name is exactly the same as the type name
                if pred_name == type_name:
                    invalid_predicates.append(pred_name)

        if invalid_predicates:
            feedback_msg = "ERROR: The following predicate(s) have the same name(s) as existing object types:"
            for pred_i, pred_name in enumerate(invalid_predicates):
                feedback_msg += f"\n{pred_i + 1}. {pred_name}"
            feedback_msg += "\nPlease rename these predicates."
            return False, feedback_msg

        feedback_msg = "PASS: All predicate names are unique to object type names"
        return True, feedback_msg

    def validate_duplicate_predicates(
        self, curr_predicates: list[Predicate], new_predicates: list[Predicate]
    ) -> tuple[bool, str]:
        """Checks if predicates have the same name but different parameters"""

        curr_pred_dict = {pred["name"].lower(): pred for pred in curr_predicates}

        duplicated_predicates = list()
        for new_pred in new_predicates:
            # check if the name is already used
            if new_pred["name"].lower() in curr_pred_dict:

                curr = curr_pred_dict[new_pred["name"].lower()]

                if len(curr["params"]) != len(new_pred["params"]) or any(
                    [t1 != t2 for t1, t2 in zip(curr["params"], new_pred["params"])]
                ):
                    # if the params are the same, then it's not a problem
                    duplicated_predicates.append(
                        (
                            new_pred["raw"],
                            curr_pred_dict[new_pred["name"].lower()]["raw"],
                        )
                    )
        if len(duplicated_predicates) > 0:
            feedback_msg = f"The following predicate(s) have the same name(s) as existing predicate(s):"
            for pred_i, duplicated_pred_info in enumerate(duplicated_predicates):
                new_pred_full, existing_pred_full = duplicated_pred_info
                feedback_msg += f'\n{pred_i + 1}. {new_pred_full.replace(":", ",")}; existing predicate with the same name: {existing_pred_full.replace(":", ",")}'
            feedback_msg += "\n\nYou should reuse existing predicates whenever possible. If you are reusing existing predicate(s), you shouldn't list them under 'New Predicates'. If existing predicates are not enough and you are devising new predicate(s), please use names that are different from existing ones."
            feedback_msg += "\n\nPlease revise the PDDL model to fix this error.\n\n"
            feedback_msg += "Parameters:"
            return False, feedback_msg

        feedback_msg = "PASS: All predicates are unique to each other."
        return True, feedback_msg

    def validate_format_predicates(
        self, predicates: list[dict], types: dict[str, str]
    ) -> tuple[bool, str]:
        """Checks for any PDDL syntax found within predicates"""

        all_invalid_params = []

        for pred in predicates:
            pred_def = pred["raw"].split(": ")[0]
            pred_def = pred_def.strip(" ()`")  # discard parentheses and similar
            split_predicate = pred_def.split(" ")[1:]  # discard the predicate name
            split_predicate = [e for e in split_predicate if e != ""]

            for i, p in enumerate(split_predicate):
                if i % 3 == 0:
                    if "?" not in p:
                        feedback_msg = f"There are syntax errors in the definition of the new predicate {pred_def}. Check for any missing '?' variables, or missing type declarations. Please revise its definition and output the entire PDDL action model again. Note that you need to strictly follow the syntax of PDDL."
                        return False, feedback_msg
                    else:
                        if (
                            i + 1 >= len(split_predicate)
                            or split_predicate[i + 1] != "-"
                        ):
                            feedback_msg = f"There are syntax errors in the definition of the new predicate {pred_def}. Please revise its definition and output the entire PDDL action model again. Note that you need to define the object type of each parameter and strictly follow the syntax of PDDL."
                            return False, feedback_msg

                        if i + 2 >= len(split_predicate):
                            feedback_msg = f"There are syntax errors in the definition of the new predicate {pred_def}. Please revise its definition and output the entire PDDL action model again. Note that you need to define the object type of each parameter and strictly follow the syntax of PDDL."
                            return False, feedback_msg

                        param_obj_type = split_predicate[i + 2].lower()

                        # Extract the base type names from the keys in types
                        valid_types = {
                            type_key.split(" - ")[0].strip().lower()
                            for type_key in types.keys()
                        }

                        # Check if the parameter object type is in the set of valid types
                        if param_obj_type not in valid_types:
                            all_invalid_params.append((param_obj_type, p, pred_def))

        if all_invalid_params:
            feedback_msg = "There are invalid object types in the predicates:"
            for param_obj_type, p, pred_def in all_invalid_params:
                feedback_msg += f"\n- `{param_obj_type}` for the parameter `{p}` in the definition of the predicate `{pred_def}` not found in types: {valid_types}."
            feedback_msg += "\nPlease revise these definitions and output the entire PDDL action model again."
            return False, feedback_msg

        feedback_msg = "PASS: All predicates are formatted correctly."
        return True, feedback_msg

    def validate_pddl_usage_predicates(
        self,
        pddl: str,
        predicates: list[Predicate],
        action_params: list[str],
        types: dict[str, str],
        part="preconditions",
    ) -> tuple[bool, str]:
        """
        This function checks three types of errors:
            - (i) check if the num of params given matches the num of params in predicate definition
            - (ii) check if there is any param that is not listed under `Parameters:`
            - (iii) check if the param type matches that in the predicate definition
        """

        def get_ordinal_suffix(_num):
            return (
                {1: "st", 2: "nd", 3: "rd"}.get(_num % 10, "th")
                if _num not in (11, 12, 13)
                else "th"
            )

        pred_names = {predicates[i]["name"]: i for i in range(len(predicates))}
        pddl_elems = [e for e in pddl.split(" ") if e != ""]

        idx = 0
        while idx < len(pddl_elems):
            if pddl_elems[idx] == "(" and idx + 1 < len(pddl_elems):
                if pddl_elems[idx + 1] in pred_names:

                    curr_pred_name = pddl_elems[idx + 1]
                    curr_pred_params = list()
                    target_pred_info = predicates[pred_names[curr_pred_name]]

                    # read params
                    idx += 2
                    while idx < len(pddl_elems) and pddl_elems[idx] != ")":
                        curr_pred_params.append(pddl_elems[idx])
                        idx += 1
                    # (i) check if the num of params are correct
                    n_expected_param = len(target_pred_info["params"])
                    if n_expected_param != len(curr_pred_params):

                        feedback_msg = f'In the {part}, the predicate `{curr_pred_name}` requires {n_expected_param} parameters but {len(curr_pred_params)} parameters were provided. Object type should not be declared in the {part}, but just the variable. For example, "(drive ?a ?from)" does not contain its object types, just variables. Do not change the predicates. Please revise the PDDL model to fix this error.'
                        return False, feedback_msg

                    # (ii) check if there is any unknown param
                    for curr_param in curr_pred_params:

                        if curr_param not in action_params[0]:
                            feedback_msg = f"In the {part} and in the predicate `{curr_pred_name}`, there is an unknown parameter `{curr_param}`. You should define all parameters (i.e., name and type) under the `### Action Parameters` list. Please revise the PDDL model to fix this error (and other potentially similar errors)."
                            return False, feedback_msg

                    # (iii) check if the object types are correct
                    target_param_types = [
                        target_pred_info["params"][t_p]
                        for t_p in target_pred_info["params"]
                    ]
                    for param_idx, target_type in enumerate(target_param_types):
                        curr_param = curr_pred_params[param_idx]
                        claimed_type = action_params[0][curr_param]

                        if not self.validate_type(target_type, claimed_type, types):
                            feedback_msg = f"There is a syntax error in the {part.lower()}, the {param_idx+1}-{get_ordinal_suffix(param_idx+1)} parameter of `{curr_pred_name}` should be a `{target_type}` but a `{claimed_type}` was given. Please use the correct predicate or devise new one(s) if needed (but note that you should use existing predicates as much as possible)."
                            return False, feedback_msg
            idx += 1

        feedback_msg = "PASS: all correct use of predicates."
        return True, feedback_msg

    def validate_usage_predicates(
        self, llm_response: str, curr_predicates: list[Predicate], types: dict[str, str]
    ):
        """
        This function performs very basic check over whether the predicates are used in a valid way.
            This check should be performed at the end.
        """

        # parse predicates
        new_predicates = parse_new_predicates(llm_response)
        curr_predicates.extend(new_predicates)
        curr_predicates = parse_predicates(curr_predicates)

        # get action params
        params_info = parse_params(llm_response)

        # check preconditions
        precond_str = llm_response.split("Preconditions")[1].split("```\n")[1].strip()
        precond_str = (
            precond_str.replace("\n", " ").replace("(", " ( ").replace(")", " ) ")
        )

        validation_info = self.validate_pddl_usage_predicates(
            precond_str, curr_predicates, params_info, types, part="preconditions"
        )
        if not validation_info[0]:
            return validation_info

        # check effects
        if llm_response.split("Effects")[1].count("```\n") < 2:
            return True, "invalid_predicate_usage"
        eff_str = llm_response.split("Effects")[1].split("```\n")[1].strip()
        eff_str = eff_str.replace("\n", " ").replace("(", " ( ").replace(")", " ) ")
        return self.validate_pddl_usage_predicates(
            eff_str, curr_predicates, params_info, types, part="effects"
        )

    def validate_overflow_predicates(
        self, llm_response: str, limit: int
    ) -> tuple[bool, str]:
        """
        Checks if LLM output contains too many predicates in precondition/effects (based on users assigned limit)
        """
        assert "\nPreconditions:" in llm_response, llm_response
        precond_str = (
            llm_response.split("\nPreconditions:")[1].split("```\n")[1].strip()
        )
        if len(precond_str.split("\n")) > limit:
            feedback_msg = f"FAIL: You seem to have generated an action model with an unusually long list of preconditions. Please include only the relevant preconditions/effects and keep the action model concise.\n\nParameters:"
            return False, feedback_msg

        eff_str = llm_response.split("Effects")[1].split("```\n")[1].strip()
        if len(eff_str.split("\n")) > limit:
            feedback_msg = f"FAIL: You seem to have generated an action model with an unusually long list of effects. Please include only the relevant preconditions/effects and keep the action model concise.\n\nParameters:"
            return False, feedback_msg

        feedback_msg = "PASS: predicate output is fine."
        return True, feedback_msg

    def validate_task_objects(
        self, objects: dict[str, str], types: dict[str, str]
    ) -> tuple[bool, str]:
        """
        Parameters:
            - objects (dict[str,str]): a dictionary of the task objects.
            - types (dict[str,str]): a dictionary of the domain types.

        Returns:
            - check, feedback_msg (bool, str)

        Checks following cases:
            (i) if object type is the same as type
            (ii) if object name is the same as type
        """

        valid = True
        feedback_msgs = []

        for obj_name, obj_type in objects.items():
            obj_type_found = False

            for type_key in types.keys():
                current_type, parent_type = type_key.split(" - ")

                # checks if obj_type is found in types
                if obj_type == current_type or obj_type == parent_type:
                    obj_type_found = True

                # checks if obj_name matches either current_type or parent_type
                if obj_name == current_type:
                    feedback_msgs.append(
                        f"ERROR: Object variable '{obj_name}' matches the type name '{current_type}', change it to be unique from types: {types.keys()}"
                    )
                    valid = False
                    break
                if obj_name == parent_type:
                    feedback_msgs.append(
                        f"ERROR: Object variable '{obj_name}' matches the type name '{parent_type}', change it to be unique from types: {types.keys()}"
                    )
                    valid = False
                    break

            # clause that checks if obj_type is found in types
            if not obj_type_found:
                feedback_msgs.append(
                    f"ERROR: Object variable '{obj_name}' has an invalid type '{obj_type}' not found in types: {types.keys()}"
                )
                valid = False

        feedback_msg = (
            "\n".join(feedback_msgs) if not valid else "PASS: all objects are valid."
        )

        return valid, feedback_msg

    def validate_task_states(
        self,
        states: list[dict[str, str]],
        objects: dict[str, str],
        predicates: list[Predicate],
        state_type: str = "initial",
    ) -> tuple[bool, str]:
        """
        Parameters:
            - states (list[dict[str,str]]): a list of dictionaries of the state states.
            - parameters (OrderedDict): parameters of the current action.
            - types (dict[str,str]): a dictionary of the domain types.

        Returns:
            - check, feedback_msg (bool, str)

        Checks following cases:
            (i) if predicates in states are found in predicates in domain
            (ii) if object variables in states are found in task object list
        """

        valid = True
        feedback_msgs = []

        # loop through each state
        for state in states:

            # (i) check if predicates in states are found in predicates in domain
            matched_preds = False
            state_name = state["name"]  # retrieve predicate name from state

            # loop through each predicate name from domain
            for pred in predicates:
                # check if predicate in state is found in predicate domain
                if state_name == pred["name"]:
                    matched_preds = True

            # if no matches, then that state is missusing a predicate - not found in domain
            if matched_preds == False:
                feedback_msgs.append(
                    f"ERROR: In the {state_type} state, '({state['name']} {' '.join(state['params'])})' contains '{state_name}' predicate, which is not found in {[p['name'] for p in predicates]}, predicate in state is missused."
                )
                valid = False

            # (ii) check if object variables in states are found in task object list
            state_params = state["params"]  # retrieve variables from state

            # loop through each parameter in current state
            for state_p in state_params:

                matched_params = False
                for obj_name, obj_type in objects.items():
                    # check if parameter is found in object names
                    if state_p == obj_name:
                        matched_params = True

                if matched_params == False:
                    feedback_msgs.append(
                        f"ERROR: In the {state_type} state, '({state['name']} {' '.join(state['params'])})' contains parameter '{state_p}' not found in '{objects.keys()}'."
                    )
                    valid = False

        feedback_msg = (
            "\n".join(feedback_msgs) if not valid else "PASS: all objects are valid."
        )

        return valid, feedback_msg

    def validate_header(self, llm_response: str):
        """Checks if domain headers and formatted code block syntax are found in LLM output"""

        for header in ["Parameters", "Preconditions", "Effects", "New Predicates"]:
            if header not in llm_response:
                feedback_msg = f"FAIL: The header `{header}` is missing in the PDDL model. Please include the header `{header}` in the PDDL model."
                return False, feedback_msg
        for header in ["Parameters", "Preconditions", "Effects"]:
            if llm_response.split(f"{header}")[1].split("##")[0].count("```\n") < 2:
                feedback_msg = f'FAIL: The header `{header}` is missing in the formalised code block. Please include a "```" section in the {header} section.'
                return False, feedback_msg

        feedback_msg = "PASS: headers are identified properly in LLM output."
        return True, feedback_msg

    def validate_unsupported_keywords(
        self, llm_response: str, unsupported_keywords: list[str]
    ) -> tuple[bool, str]:
        """Checks whether PDDL model uses unsupported logic keywords"""

        for key in unsupported_keywords:
            if f"{key}" in llm_response:
                feedback_msg = (
                    f"ERROR: The precondition or effect contains the keyword {key}."
                )
                return False, feedback_msg

        feedback_msg = "PASS: Unsupported keywords not found in PDDL model."
        return True, feedback_msg

    def validate_keyword_usage(self, llm_response: str):
        """Checks if action effects uses unsupported universal condition keywords"""

        if not "Action Effects" in llm_response:
            feedback_msg = "PASS"
            return True, feedback_msg
        heading = llm_response.split("Action Effects")[1].split("```\n")[1].strip()
        for keyword in ["forall", "exists", "if "]:
            if keyword in heading:
                feedback_msg = (
                    f"The keyword `{keyword}` is not supported in the action effects."
                )
                return False, feedback_msg

        feedback_msg = "PASS: unsupported keywords are not found in the action effects."
        return True, feedback_msg

    def validate_new_action_creation(self, llm_response: str) -> tuple[bool, str]:
        """Checks if the LLM attempts to create a new action (so two or more actions defined in the same response)"""

        if (
            llm_response.count("## Action Parameters") > 1
            or llm_response.count("## Preconditions") > 1
            or llm_response.count("## Effects") > 1
            or llm_response.count("## New Predicates") > 1
        ):
            feedback_msg = "It's not possible to create new actions at this time. Please only define the requested action."
            return False, feedback_msg

        feedback_msg = "PASS: no new actions created"
        return True, feedback_msg

    def validate_type(self, target_type, claimed_type, types):
        """
        Check if the claimed_type is valid for the target_type according to the type hierarchy.

        Parameters:
            - target_type (str): The type that is expected for the parameter.
            - claimed_type (str): The type that is provided in the PDDL.
            - types (dict[str, str]): A dictionary mapping subtypes to their supertypes.

        Returns:
            - bool: True if claimed_type is valid, False otherwise.
        """
        # Check if the claimed type matches the target type
        if claimed_type == target_type:
            return True

        # Iterate through the types hierarchy to check if claimed_type is a subtype of target_type
        current_type = claimed_type

        # Extract all types from the keys in the types dictionary
        all_types = set()
        type_hierarchy = {}
        for key in types.keys():
            main_type, *subtype = key.split(" - ")
            all_types.add(main_type.strip())
            if subtype:
                all_types.add(subtype[0].strip())
                type_hierarchy[subtype[0].strip()] = main_type.strip()

        while current_type in all_types:
            # find the key that starts with the current type

            parent_type_entry = next(
                (k for k in types.keys() if k.startswith(f"{current_type} - ")), None
            )

            if parent_type_entry:
                # extract the parent type from the key
                super_type = parent_type_entry.split(" - ")[1].strip()

                if super_type == target_type:
                    return True
                current_type = super_type
            else:
                break

        return False
